#
# This file is part of python sample package for the BROADWAY2 package.
#
# MIT License

# Copyright (c) 2013-2020 FibreCode GmbH <info@fibrecode.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Important note:
# Starting with broadway2 V2.3.0 there is a split to stay on pure SMI/SPI access
# at broadway2-API.
# Means, all files/classes regarding register/property encoding to all Transceivers
# like NXP TJA1100/TJA1101 are no more availble on the broadway2 python installer.
# More details can be found on UM_broadway2_xxxx.pdf


import sys, os
import platform
import time, datetime

from broadway2 import Broadway, SmiAccess, PhyAccess
from broadway2 import broadway_api, get_phy_access

IsSupported = {602: True,
               611: True,
               612: True,
               621: True,
               631: True,
               601: False,
               401: True
               }

def fc401_basics(base, adapter):
    toolAdapter = base.get_adapter(adapter.Inst)
        

    print('FC401 SMI-SPI-Adapter {0} found'.format(toolAdapter.SerialNumber))

    dev = base.open_control(toolAdapter.Inst)

    toolAdapter.connect(dev)

    smiAccess = SmiAccess(dev) # create spi adapter to get interface

    res = smiAccess.detect2()

    if len(res) == 0:
        print('... nothing connected on SMI')
    else:
        print('SMI-Detected: {0}'.format(len(res)))

    for phy in res:
        phyDetail = res[phy]
        print('Phy found on {0}. ID1: {1:04X} ID2: {2:04X} Clause: {3}'.format(phy, phyDetail['ID1'], phyDetail['ID2'], phyDetail['Clause']))

    # now let's read some basics


        phyA = get_phy_access(toolAdapter, phy, phyDetail['Clause'])

        # try clause 45

        if phyA.supportsClause45:

            reg2 = phyA.read_dev_reg(0x1E, 2)
            reg3 = phyA.read_dev_reg(0x1E, 3)

            print('{0} / C45: {1:04X} {2:04X}'.format(phy, reg2, reg3))

        if phyA.supportsClause22:
            
            reg2 = phyA.read_reg(2)
            reg3 = phyA.read_reg(3)

            print('{0} / C22: {1:04X} {2:04X}'.format(phy, reg2, reg3))

    smiAccess.disconnect()

    base.close(dev)

def fc602_basics(base, adapter):
    netAdapter = base.get_adapter(adapter.Inst)
        

    print('FC602 Net-100Base-T1-Adapter {0} found'.format(netAdapter.SerialNumber))

    dev = base.open_control(netAdapter.Inst)

    netAdapter.connect(dev)

    phy = get_phy_access(netAdapter)

    # access via API

    phyId2 = phy.read_reg(0x02)
    phyId3 = phy.read_reg(0x03)
    print('PhyID: {0:04X}.{1:04X}'.format(phyId2, phyId3))

    base.close(dev)

def fc611_basics(base, adapter):
    rawAdapter = base.get_adapter(adapter.Inst)
        

    print('FC611 Raw-100Base-T1-Adapter {0} found'.format(rawAdapter.SerialNumber))

    dev = base.open_control(rawAdapter.Inst)
    rawAdapter.connect(dev)

    phy = get_phy_access(rawAdapter)

    # access via API

    phyId2 = phy.read_reg(0x02)
    phyId3 = phy.read_reg(0x03)
    print('PhyID: {0:04X}.{1:04X}'.format(phyId2, phyId3))

    base.close(dev)

def fc612_basics(base, adapter):
    rawAdapter = base.get_adapter(adapter.Inst)
        

    print('FC612 Raw-100Base-T1-Adapter {0} found'.format(rawAdapter.SerialNumber))

    dev = base.open_control(rawAdapter.Inst)
    rawAdapter.connect(dev)

    phy = get_phy_access(rawAdapter)

    # access via API

    phyId2 = phy.read_reg(0x02)
    phyId3 = phy.read_reg(0x03)
    print('PhyID: {0:04X}.{1:04X}'.format(phyId2, phyId3))

    base.close(dev)

def fc621_basics(base, adapter):
    net10L = base.get_adapter(adapter.Inst)
        

    print('FC621 Net-10Base-T1L-Adapter {0} found'.format(net10L.SerialNumber))

    dev = base.open_control(net10L.Inst)
    net10L.connect(dev)

    phyAccess = get_phy_access(net10L, 2)
    
    phyId2 = phyAccess.read_dev_reg(0x1E, 0x02)
    phyId3 = phyAccess.read_dev_reg(0x1E, 0x03)
    print('PhyID: {0:04X}.{1:04X}'.format(phyId2, phyId3))

    base.close(dev)

def fc631_basics(base, adapter):
    raw10L = base.get_adapter(adapter.Inst)
        

    print('FC631 Raw-10Base-T1L-Adapter {0} found'.format(raw10L.SerialNumber))

    dev = base.open_control(raw10L.Inst)
    raw10L.connect(dev)

    phyAccess = get_phy_access(raw10L, 2)
    
    phyId2 = phyAccess.read_dev_reg(0x1E, 0x02)
    phyId3 = phyAccess.read_dev_reg(0x1E, 0x03)
    print('PhyID: {0:04X}.{1:04X}'.format(phyId2, phyId3))


    base.close(dev)

def main(args):
    # create adapter singleton
    funcLut = {401: fc401_basics, 602: fc602_basics, 611: fc611_basics, 612: fc612_basics, 621: fc621_basics, 631: fc631_basics}
    
    base = Broadway()


    for adapter in base.enumerate():
        try:
            supported = IsSupported[adapter.DeviceType]
        except:
            supported = False   # not in the list at all

        if supported:

            print('FWName: {0} Rev.:{1} Serial:{2} Name: {3}'.format(adapter.FWName, adapter.FWRevision, adapter.SerialNumber, adapter.FriendlyName))
            print('='*78)
       
            try:
                func = funcLut[adapter.DeviceType]        
                func(base, adapter)
            except:
                pass

            print()


if __name__ == "__main__":
    main(sys.argv)
