﻿#!python3
# broadway_api.py
#
# Broadway2 API V2.3.2
#
# This file is part of python wrapper for the BROADWAY2 package.
#
# MIT License

# Copyright (c) 2013-2022 FibreCode GmbH <info@fibrecode.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import os
import platform
import time
import threading
import ctypes


''' Broadway Python API Wrapper '''

_WinDllName = 'broadway2_api.dll'
_WinDll64Name = 'broadway2_api.dll'
_LinuxSoName = 'libbroadway2_api.so'

NULLPTR = ctypes.POINTER(ctypes.c_int)()
TSTATUS = ctypes.c_ulong
TOBJECTID = ctypes.c_int


# Initialize global Instance to Dll/so
_lib = None
LastResult = None

CBDeviceChanged = None

# T_UNICHAR needs to be handled different on Windows/Linux
# Windows: wchar_t
# Linux: char
T_UNICHAR = None

class FCBDeviceType:
    FCBDeviceTypeFC401 = 401
    FCBDeviceTypeFC602 = 602
    FCBDeviceTypeFC611 = 611
    FCBDeviceTypeFC612 = 612
    FCBDeviceTypeFC621 = 621
    FCBDeviceTypeFC631 = 631
    FCBDeviceTypeAny = -1


class FCBPinState:
    FCBPinStateInactive = 0
    FCBPinStateActive = 1

class FCBLedControlMode:
    #The LED is controlled by the device.
    FCBLedControlMode_Device = 0
    #The LED is controlled by the host through the BROADWAY API.
    FCBLedControlMode_Host = 1

class FCBLedId:
    Red=1
    Green=2
    Blue=3


class FCBSpiProtocol:
    # SPI selection for correct CPHA/CPOL settings
    SJA1105=0
    SJA1110=1

class FCBDeviceProperty:
    FCBDevicePropertyReserved = 0

    FCB_PROP_GROUP_USB = 0x100
    
    FCBDevicePropertyUsbVendorId = FCB_PROP_GROUP_USB + 1
    FCBDevicePropertyUsbProductId = FCB_PROP_GROUP_USB + 2
    FCBDevicePropertyUsbRevBcd = FCB_PROP_GROUP_USB + 3
    FCBDevicePropertyUsbSerialNumberString = FCB_PROP_GROUP_USB + 4
    FCBDevicePropertyUsbManufacturerString = FCB_PROP_GROUP_USB + 5
    FCBDevicePropertyUsbProductIdString = FCB_PROP_GROUP_USB + 6
    FCBDevicePropertyMaxInMessageSize = FCB_PROP_GROUP_USB + 7
    FCBDevicePropertyMaxOutMessageSize = FCB_PROP_GROUP_USB + 8
    FCBDevicePropertyDeviceType = FCB_PROP_GROUP_USB + 9
    

    FCB_PROP_ID_SETUP_REQ = 0x8000

    FCBDevicePropertyMacAddress = 0x0110 | FCB_PROP_ID_SETUP_REQ
    FCBDevicePropertyFwVariantName = 0x0111 | FCB_PROP_ID_SETUP_REQ
    FCBDevicePropertyFwVerString = 0x0112 | FCB_PROP_ID_SETUP_REQ
    FCBDevicePropertySiliconKey = 0x0113 | FCB_PROP_ID_SETUP_REQ
    FCBDevicePropertyPhyTypeString = 0x0114 | FCB_PROP_ID_SETUP_REQ
    FCBDevicePropertySerialNoBase = 0x0115 | FCB_PROP_ID_SETUP_REQ
    FCBDevicePropertySerialNoAddOn = 0x0116 | FCB_PROP_ID_SETUP_REQ
    FCBDevicePropertyMtpVersionReadable = 0x0117 | FCB_PROP_ID_SETUP_REQ
    FCBDevicePropertyFriendlyName = 0x0118 | FCB_PROP_ID_SETUP_REQ

    FCBDevicePropertyPhyLinkState = 0x0301 | FCB_PROP_ID_SETUP_REQ
    FCBDevicePropertyPhyLinkMode = 0x0302 | FCB_PROP_ID_SETUP_REQ
    FCBDevicePropertyReset = 0x0303 | FCB_PROP_ID_SETUP_REQ
    FCBDevicePropertyFrameInterval = 0x0304 | FCB_PROP_ID_SETUP_REQ
    FCBDevicePropertyMasterSlave = 0x0305 | FCB_PROP_ID_SETUP_REQ
    #FCBDevicePropertyTimestamp = 0x0306 | FCB_PROP_ID_SETUP_REQ   #deprecated in V3!
    FCBDevicePropertyMACAddressOverride = 0x0307 | FCB_PROP_ID_SETUP_REQ
    FCBDevicePropertyUseMACAddressOverride = 0x0308 | FCB_PROP_ID_SETUP_REQ
    # new Timstamp based on HW-MAC
    FCBDevicePropertyPTPTimestamp = 0x0309 | FCB_PROP_ID_SETUP_REQ
    FCBDevicePropertyPTPTimestampInit = 0x030A | FCB_PROP_ID_SETUP_REQ
    FCBDevicePropertyPTPTimestampUpdate = 0x030B | FCB_PROP_ID_SETUP_REQ
    FCBDevicePropertyPTPTimestampClockSpeed = 0x030C | FCB_PROP_ID_SETUP_REQ

    # new feature for TC-10 compliant Phys
    FCBDevicePropertyPhyWake = 0x030D | FCB_PROP_ID_SETUP_REQ

    # spi configuration settings
    FCBPropertySpiProtocol = 0x030E | FCB_PROP_ID_SETUP_REQ
    FCBPropertySpiBusSpeed = 0x030F | FCB_PROP_ID_SETUP_REQ

    # External Trigger-In on UART-RX for Raw and SMI/SPI Tool Adapters
    FCBPropertyExternalTriggerConfig = 0x0310 | FCB_PROP_ID_SETUP_REQ
    FCBPropertyExternalTriggerStatus = 0x0311 | FCB_PROP_ID_SETUP_REQ

    # 10Base-T1L Options
    FCBPropertySupportLongDistance = 0x0312 | FCB_PROP_ID_SETUP_REQ
    FCBPropertyEnableLongDistance = 0x0313 | FCB_PROP_ID_SETUP_REQ

    # Tx Shaping
    FCBPropetyMinTxFrameInterval = 0x0314 | FCB_PROP_ID_SETUP_REQ

    # Trigger Output on UART-TX -> Trig_Out
    FCBPropertyTriggerOutputConfig = 0x0315 | FCB_PROP_ID_SETUP_REQ
    FCBPropertyTriggerOutputStatus = 0x0316 | FCB_PROP_ID_SETUP_REQ
    FCBPropertyForceTrigger = 0x0317 | FCB_PROP_ID_SETUP_REQ
    FCBPropertySmiInterfacePassive = 0x0318 | FCB_PROP_ID_SETUP_REQ
    FCBPropertySpiInterfacePassive = 0x0319 | FCB_PROP_ID_SETUP_REQ

    # Statistic
    FCBPropertyQueryFrameStatisticsV1 = 0x0600 | FCB_PROP_ID_SETUP_REQ
    FCBPropertyQueryAndClearFrameStatisticsV1 = 0x0601 | FCB_PROP_ID_SETUP_REQ
    

class FCBPinIds:
    Invalid = 0
    BROADWAY_PIN_GROUP_PHY = 0x1000
    PhyReset = BROADWAY_PIN_GROUP_PHY + 1
    PhyWake = BROADWAY_PIN_GROUP_PHY + 2
    PhyEnable = BROADWAY_PIN_GROUP_PHY + 3
    PhyInterrupt = BROADWAY_PIN_GROUP_PHY + 4
    PhyLED = BROADWAY_PIN_GROUP_PHY + 5
    PhyInh = BROADWAY_PIN_GROUP_PHY + 6

class FCBPhyMode:
    FCBMasterMode=0
    FCBSlaveMode=1

class FCBTxResources:

    FCBTxResourcesOneHalf=1


class RawRxHeaderType:
    NoHeader = 0
    Type1 = 1   # Type1 is deprecated since API 2.12 to get full nanosec from LPC MAC-timestamp
    Type2 = 2

class RawRxFrameType:
    FCBRawFrameTypeUnknown = 0
    FCBRawFrameTypeData = 1
    FCBRawFrameTypeIdle = 2
    FCBRawFrameTypeExternalTrigger = 3


class FCBPropertyFlags:
    FCB_PROP_FLAG_SET_PERSISTENT=1

FCB_PTP_TIMESTAMP_INVALID = 0xFFFFFFFFFFFFFFFF

class FrameStatsV1:
    def __init__(self, binArray):
        """create all values from binary struct / 12 Bytes Timestamp + 12*8 = 108 Bytes in V1"""
        secs = int.from_bytes(binArray[0:8], 'little')

        if secs == FCB_PTP_TIMESTAMP_INVALID:   # Node-Sticks has no precise hw-timebase
            tm = time.time() # use Host-Time
            self.TimeStampSec = int(tm)
            self.TimeStampNano = round((tm%1)*1e9)
        else:
            self.TimeStampSec = secs
            self.TimeStampNano = int.from_bytes(binArray[8:12], 'little')

        self.TxFramesTotal = int.from_bytes(binArray[12:20], 'little')          # 0
        self.TxBytesOk = int.from_bytes(binArray[20:28], 'little')              # 1
        self.RxFramesTotal = int.from_bytes(binArray[28:36], 'little')          # 2
        self.RxFramesOk = int.from_bytes(binArray[36:44], 'little')             # 3
        self.RxBytesOk = int.from_bytes(binArray[44:52], 'little')              # 4

        self.RxFramesDefective = int.from_bytes(binArray[52:60], 'little')      # 5
        self.RxFramesCrcError = int.from_bytes(binArray[60:68], 'little')       # 6
        self.RxFramesLengthError = int.from_bytes(binArray[68:76], 'little')    # 7
        
        self.RxFramesRejected = int.from_bytes(binArray[76:84], 'little')       # 8
        self.RxFramesUsbCongestion = int.from_bytes(binArray[84:92], 'little')  # 9

        self.LinkUpEvents = int.from_bytes(binArray[92:100], 'little')          # 10
        self.LinkDownEvents = int.from_bytes(binArray[100:108], 'little')       # 11

    def get_total_frames(self):
        return (self.TxFramesTotal, self.RxFramesTotal)

    def get_crc_error_count(self):
        return self.RxFramesCrcError

    def get_linkup_count(self):
        return self.LinkUpEvents

    def get_linkdown_count(self):
        return self.LinkDownEvents

#void
#FCBASE_CALL
#FCB_DeviceChangeCallback(
# void* callbackContext  
# FCB_DeviceEvent eventType,
# ) 

CB_DeviceChangeEvent = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.c_int)


# Number of Bytes per HID receive/transmit function
MAX_RX_BUFFER=1024
MAX_TX_BUFFER=1024



def initialize(lib_path=''):
    """Initialize the module and Library.
    
    Load the library, the error codes and the functions and call the library's 
    Initialize function.
    
    """
    global T_UNICHAR
    try:
        # Load the library, error codes and functions.
        res = __load_library(lib_path)
        #__load_error_codes()
        if res:
            __load_functions()

           # verify correct char-size
            res = get_unichar_size()
            if res == 2:
                T_UNICHAR = ctypes.c_wchar
            elif res == 1:
                T_UNICHAR = ctypes.c_char
            else:
                raise

        # private functions
        
    except LibLoadError:
        raise
            
def uninitialize():
    """Uninitialize the library to release the device handles.
    
    This function gets called automatically at exit if the libraries
    initialize() function got called sucessfully.
    
    Raises:
        LibInitError: Raised if library fails to be uninitialized.
        
    @see hid2inic_api.initialize()

    @ingroup hid2inic_api
    
    """
    global _lib
    if _lib:
        pass
                
                
def is_os_64bit():
    """helper function to detect running OS."""
    return platform.machine().endswith('64')

def is_python_64bit():
    """helper function to detect running python 32 or 64. In case of loading Dlls, we need right Dlls to load"""
    return platform.sys.maxsize > 2**32

def __load_library(lib_path=''):
    """
    Initialize python wrapper to load Dll/so. If no path is given, default path will be used.
    """
    global _lib
    global char_type
    # verify target system
    is_64bits = is_python_64bit()
    
    try:
        if platform.platform().startswith('Win'):
            T_UNICHAR = ctypes.c_wchar
            # check if you are on a 32bit or 64bit architecture    
            if is_64bits:
                _lib = ctypes.WinDLL(lib_path + _WinDll64Name)
            else:
                _lib = ctypes.WinDLL(lib_path + _WinDllName)
            
        else:
            T_UNICHAR = ctypes.c_char
            
            _lib = ctypes.CDLL(lib_path + _LinuxSoName)
            
    except OSError:
        lib_load_fail = True
    else:
        lib_load_fail = False
    finally:
        if lib_load_fail:
            raise LibLoadError(lib_path)

    return True
            

def __load_error_codes():
    """Load the error codes from the library
        
    Raises:
        LibInitError: Raised if symbol could not be found in library.
        
    """
    try:
        global NO_ERROR
        NO_ERROR = ctypes.c_int.in_dll(_lib, 'NO__ERROR').value
        global NOT_INITIALIZED_ERROR
        NOT_INITIALIZED_ERROR = ctypes.c_int.in_dll(_lib,'NOT_INITIALIZED_ERROR').value
        global DEVICE_ERROR
        DEVICE_ERROR = ctypes.c_int.in_dll(_lib,'DEVICE_ERROR').value
        global INVALID_ARGUMENT_ERROR
        INVALID_ARGUMENT_ERROR = ctypes.c_int.in_dll(_lib,'INVALID_ARGUMENT_ERROR').value
        global POINTER_ERROR
        POINTER_ERROR = ctypes.c_int.in_dll(_lib,'POINTER_ERROR').value
        global OUT_OF_MEMORY_ERROR
        OUT_OF_MEMORY_ERROR = ctypes.c_int.in_dll(_lib,'OUT_OF_MEMORY_ERROR').value
        global UNKNOWN_ERROR
        UNKNOWN_ERROR = ctypes.c_int.in_dll(_lib,'UNKNOWN_ERROR').value
        global NOT_IMPLEMENTED
        NOT_IMPLEMENTED = ctypes.c_int.in_dll(_lib,'NOT_IMPLEMENTED').value
    except ValueError as err:
        error_msg = err.args[0]
    else:
        error_msg = ''
    finally:
        if error_msg:
            raise LibInitError(error_msg, _lib._name)
        
        
def __load_functions():
    """Load and initialize all functions from the library.
        
    Raises:
        LibInitError: Raised if function could not be found in library.
        
    """
    global _lib
    global char_type
    
    try:
        error_msg = ''
        
        # int FCB_GetApiVersion()
        global FCB_GetApiVersion
        FCB_GetApiVersion = _lib['FCB_GetApiVersion']
        FCB_GetApiVersion.restype = ctypes.c_uint32
        FCB_GetApiVersion.argtypes = []
        
        # int FCB_CheckApiVersion(int major, int minor)
        global FCB_CheckApiVersion
        FCB_CheckApiVersion = _lib['FCB_CheckApiVersion']
        FCB_CheckApiVersion.restype = ctypes.c_uint32
        FCB_CheckApiVersion.argtypes = [ctypes.c_uint32,ctypes.c_uint32]

        global FCB_GetUniCharSize
        FCB_GetUniCharSize = _lib['FCB_GetUniCharSize']
        FCB_GetUniCharSize.restype = ctypes.c_uint32
        FCB_GetUniCharSize.argtypes = []

        global FCB_GetErrorText
        FCB_GetErrorText = _lib['FCB_GetErrorText']
        FCB_GetErrorText.restype = ctypes.c_char_p
        FCB_GetErrorText.argtypes = [ctypes.c_uint32]

        # int FCB_RegisterDeviceChangeCallback()
        global FCB_RegisterDeviceChangeCallback
        FCB_RegisterDeviceChangeCallback = _lib['FCB_RegisterDeviceChangeCallback']
        FCB_RegisterDeviceChangeCallback.restype = TSTATUS
        FCB_RegisterDeviceChangeCallback.argtypes = [ctypes.c_int, CB_DeviceChangeEvent, ctypes.c_int, ctypes.POINTER(ctypes.c_uint32)]

        # UPI_UnregisterDeviceChangeCallback()
        global FCB_UnregisterDeviceChangeCallback
        FCB_UnregisterDeviceChangeCallback = _lib['FCB_UnregisterDeviceChangeCallback']
        FCB_UnregisterDeviceChangeCallback.restype = TSTATUS
        FCB_UnregisterDeviceChangeCallback.argtypes = [ctypes.c_uint]

        # int FCB_EnumerateDevices(FCBDeviceType deviceType, FCBEnumerationHandle* enumerationHandle, unsigned int* deviceCount);
        global FCB_EnumerateDevices
        FCB_EnumerateDevices = _lib['FCB_EnumerateDevices']
        FCB_EnumerateDevices.restype = ctypes.c_uint32
        FCB_EnumerateDevices.argtypes = [ctypes.c_int, ctypes.POINTER(ctypes.c_uint), ctypes.POINTER(ctypes.c_uint32)]
         
        # int FCB_CloseEnumeration(FCBEnumerationHandle enumerationHandle);
        global FCB_CloseEnumeration
        FCB_CloseEnumeration = _lib['FCB_CloseEnumeration']
        FCB_CloseEnumeration.restype = ctypes.c_uint32
        FCB_CloseEnumeration.argtypes = [ctypes.c_uint]

         # int FCB_CompareDeviceInstance(FCBEnumerationHandle enumerationHandle, unsigned int deviceIndex, FCBDeviceHandle deviceHandle);
        global FCB_CompareDeviceInstance
        FCB_CompareDeviceInstance = _lib['FCB_CompareDeviceInstance']
        FCB_CompareDeviceInstance.restype = ctypes.c_uint32
        FCB_CompareDeviceInstance.argtypes = [ctypes.c_uint, ctypes.c_uint32, ctypes.c_uint32]

        # int FCB_OpenControlDevice(FCBEnumerationHandle enumerationHandle, unsigned int deviceIndex, FCBDeviceHandle* deviceHandle);
        global FCB_OpenControlDevice
        FCB_OpenControlDevice = _lib['FCB_OpenControlDevice']
        FCB_OpenControlDevice.restype = ctypes.c_uint32
        FCB_OpenControlDevice.argtypes = [ctypes.c_uint, ctypes.c_uint32, ctypes.POINTER(ctypes.c_uint32)]
        
        # int FCB_CloseDevice(int dev);
        global FCB_CloseDevice
        FCB_CloseDevice = _lib['FCB_CloseDevice']
        FCB_CloseDevice.restype = ctypes.c_uint32
        FCB_CloseDevice.argtypes = [ctypes.c_uint32]

        # FCB_CheckDeviceConnection(FCBDeviceHandle deviceHandle);
        global FCB_CheckDeviceConnection
        FCB_CheckDeviceConnection = _lib['FCB_CheckDeviceConnection']
        FCB_CheckDeviceConnection.restype = ctypes.c_uint32
        FCB_CheckDeviceConnection.argtypes = [ctypes.c_uint32]

        # FCB_GetDevicePropertyUint(FCBDeviceHandle deviceHandle, FCBDeviceProperty propertyId, unsigned int* propertyValue);
        global FCB_GetDevicePropertyUint
        FCB_GetDevicePropertyUint = _lib['FCB_GetDevicePropertyUint']
        FCB_GetDevicePropertyUint.restype = ctypes.c_uint32
        FCB_GetDevicePropertyUint.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.POINTER(ctypes.c_uint32)]

        # FCB_SetDevicePropertyUint(FCBDeviceHandle deviceHandle, unsigned int propertyId, unsigned int propertyValue, unsigned int flags);
        global FCB_SetDevicePropertyUint
        FCB_SetDevicePropertyUint = _lib['FCB_SetDevicePropertyUint']
        FCB_SetDevicePropertyUint.restype = ctypes.c_uint32
        FCB_SetDevicePropertyUint.argtypes = [ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint32]

        # FCB_GetDevicePropertyUint64(FCBDeviceHandle deviceHandle, FCBDeviceProperty propertyId, unsigned long long* propertyValue);
        global FCB_GetDevicePropertyUint64
        FCB_GetDevicePropertyUint64 = _lib['FCB_GetDevicePropertyUint64']
        FCB_GetDevicePropertyUint64.restype = ctypes.c_uint32
        FCB_GetDevicePropertyUint64.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.POINTER(ctypes.c_uint64)]

        # FCB_SetDevicePropertyUint64(FCBDeviceHandle deviceHandle, unsigned int propertyId, unsigned long long propertyValue, unsigned int flags);
        global FCB_SetDevicePropertyUint64
        FCB_SetDevicePropertyUint64 = _lib['FCB_SetDevicePropertyUint64']
        FCB_SetDevicePropertyUint64.restype = ctypes.c_uint32
        FCB_SetDevicePropertyUint64.argtypes = [ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint64, ctypes.c_uint32]

        # FCB_GetDevicePropertyString(FCBDeviceHandle deviceHandle, FCBDeviceProperty propertyId, T_UNICHAR* stringBuffer, unsigned int stringBufferMaxItems);
        global FCB_GetDevicePropertyString
        FCB_GetDevicePropertyString = _lib['FCB_GetDevicePropertyString']
        FCB_GetDevicePropertyString.restype = ctypes.c_uint32
        FCB_GetDevicePropertyString.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.POINTER(T_UNICHAR), ctypes.c_uint]

        # FCB_SetDevicePropertyString(FCBDeviceHandle deviceHandle, FCBDeviceProperty propertyId, T_UNICHAR* stringBuffer, unsigned int stringBufferMaxItems, unsigned int flags);
        global FCB_SetDevicePropertyString
        FCB_SetDevicePropertyString = _lib['FCB_SetDevicePropertyString']
        FCB_SetDevicePropertyString.restype = ctypes.c_uint32
        FCB_SetDevicePropertyString.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.POINTER(T_UNICHAR), ctypes.c_uint32]

        # FCB_GetDevicePropertyBinary(FCBDeviceHandle deviceHandle, FCBDeviceProperty propertyId, void* buffer, unsigned int bufferSize, unsigned int* bytesReturned);
        global FCB_GetDevicePropertyBinary
        FCB_GetDevicePropertyBinary = _lib['FCB_GetDevicePropertyBinary']
        FCB_GetDevicePropertyBinary.restype = ctypes.c_uint32
        FCB_GetDevicePropertyBinary.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.c_void_p, ctypes.c_uint, ctypes.POINTER(ctypes.c_uint)]

        # FCB_SetDevicePropertyBinary(FCBDeviceHandle deviceHandle, FCBDeviceProperty propertyId, void* propertyData, unsigned int propertyDataSize, unsigned int flags);
        global FCB_SetDevicePropertyBinary
        FCB_SetDevicePropertyBinary = _lib['FCB_SetDevicePropertyBinary']
        FCB_SetDevicePropertyBinary.restype = ctypes.c_uint32
        FCB_SetDevicePropertyBinary.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.c_void_p, ctypes.c_uint, ctypes.c_uint32]

        # FCB_QueryPin(FCBDeviceHandle deviceHandle, unsigned int pinId, FCBPinState* pinState);
        global FCB_QueryPin
        FCB_QueryPin = _lib['FCB_QueryPin']
        FCB_QueryPin.restype = ctypes.c_uint32
        FCB_QueryPin.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.POINTER(ctypes.c_uint)]

         # FCB_SetPin(FCBDeviceHandle deviceHandle, unsigned int pinId, FCBPinState pinState);
        global FCB_SetPin
        FCB_SetPin = _lib['FCB_SetPin']
        FCB_SetPin.restype = ctypes.c_uint32
        FCB_SetPin.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.c_uint]

        # FCB_GetLedMode(FCBDeviceHandle deviceHandle, unsigned int ledId, FCBLedControlMode* controlMode, int* state, int* onTime, int* offTime);
        global FCB_GetLedMode
        FCB_GetLedMode = _lib['FCB_GetLedMode']
        FCB_GetLedMode.restype = ctypes.c_uint32
        FCB_GetLedMode.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.POINTER(ctypes.c_uint), ctypes.POINTER(ctypes.c_uint), ctypes.POINTER(ctypes.c_uint), ctypes.POINTER(ctypes.c_uint)]

        # FCB_SetLedMode(FCBDeviceHandle deviceHandle, unsigned int ledId, FCBLedControlMode controlMode, int state, int onTime, int offTime);
        global FCB_SetLedMode
        FCB_SetLedMode = _lib['FCB_SetLedMode']
        FCB_SetLedMode.restype = ctypes.c_uint32
        FCB_SetLedMode.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.c_uint, ctypes.c_uint, ctypes.c_uint, ctypes.c_uint]


        # ------------ PHY API -------------
        # PHY_ReadRegister(FCBDeviceHandle deviceHandle, unsigned int registerAddress, unsigned int* registerValue);
        global PHY_ReadRegister
        PHY_ReadRegister = _lib['PHY_ReadRegister']
        PHY_ReadRegister.restype = ctypes.c_uint32
        PHY_ReadRegister.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.POINTER(ctypes.c_uint)]

        # PHY_WriteRegister(FCBDeviceHandle deviceHandle, unsigned int registerAddress, unsigned int* registerValue);
        global PHY_WriteRegister
        PHY_WriteRegister = _lib['PHY_WriteRegister']
        PHY_WriteRegister.restype = ctypes.c_uint32
        PHY_WriteRegister.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.c_uint]

        # ------------ PHY API Extended for FC401 -------------
        # PHY_ReadRegisterEx(FCBDeviceHandle deviceHandle, unsigned int registerAddress, unsigned int* registerValue);
        global PHY_ReadRegisterEx
        PHY_ReadRegisterEx = _lib['PHY_ReadRegisterEx']
        PHY_ReadRegisterEx.restype = ctypes.c_uint32
        PHY_ReadRegisterEx.argtypes = [ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint, ctypes.POINTER(ctypes.c_uint)]

        # PHY_WriteRegisterEx(FCBDeviceHandle deviceHandle, unsigned int registerAddress, unsigned int* registerValue);
        global PHY_WriteRegisterEx
        PHY_WriteRegisterEx = _lib['PHY_WriteRegisterEx']
        PHY_WriteRegisterEx.restype = ctypes.c_uint32
        PHY_WriteRegisterEx.argtypes = [ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint, ctypes.c_uint]
        
        # ------------ PHY API Extended for FC621/FC631 -------------
        # PHY_ReadRegister45Ex(FCBDeviceHandle deviceHandle, unsigned int portAddress, unsigned int deviceAddress, unsigned int registerAddress, unsigned int* registerValue);
        global PHY_ReadRegister45Ex
        PHY_ReadRegister45Ex = _lib['PHY_ReadRegister45Ex']
        PHY_ReadRegister45Ex.restype = ctypes.c_uint32
        PHY_ReadRegister45Ex.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.c_uint, ctypes.c_uint, ctypes.POINTER(ctypes.c_uint)]

        # PHY_WriteRegister45Ex(FCBDeviceHandle deviceHandle, unsigned int portAddress, unsigned int deviceAddress, unsigned int registerAddress, unsigned int registerValue);
        global PHY_WriteRegister45Ex
        PHY_WriteRegister45Ex = _lib['PHY_WriteRegister45Ex']
        PHY_WriteRegister45Ex.restype = ctypes.c_uint32
        PHY_WriteRegister45Ex.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.c_uint, ctypes.c_uint, ctypes.c_uint]
        
        # PHY_ReadMultRegisters45Ex(FCBDeviceHandle deviceHandle, unsigned int portAddress, unsigned int deviceAddress, unsigned int registerAddress, unsigned int numberOfRegisters, unsigned int* registerValues);
        global PHY_ReadMultRegisters45Ex
        PHY_ReadMultRegisters45Ex = _lib['PHY_ReadMultRegisters45Ex']
        PHY_ReadMultRegisters45Ex.restype = ctypes.c_uint32
        PHY_ReadMultRegisters45Ex.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.c_uint, ctypes.c_uint, ctypes.c_uint, ctypes.c_void_p]

        # ----------- RAW Ethernet API
        # FCB_OpenRawTxDevice(FCBEnumerationHandle enumerationHandle, unsigned int deviceIndex, unsigned int bufferCount, FCBRawHeaderType headerType, FCBDeviceHandle* deviceHandle, unsigned int flags);
        global FCB_OpenRawTxDevice
        FCB_OpenRawTxDevice = _lib['FCB_OpenRawTxDevice']
        FCB_OpenRawTxDevice.restype = ctypes.c_uint32
        FCB_OpenRawTxDevice.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.c_uint, ctypes.c_uint, ctypes.POINTER(ctypes.c_uint), ctypes.c_uint]

        # FCB_WaitForTxResources(FCBDeviceHandle deviceHandle, unsigned int flags, TLTimeoutInterval waitTimeout);
        global FCB_WaitForTxResources
        FCB_WaitForTxResources = _lib['FCB_WaitForTxResources']
        FCB_WaitForTxResources.restype = ctypes.c_uint32
        FCB_WaitForTxResources.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.c_int]

        # FCB_CheckSubmitRawFrame(FCBDeviceHandle deviceHandle, unsigned int frameDataSize);
        global FCB_CheckSubmitRawFrame
        FCB_CheckSubmitRawFrame = _lib['FCB_CheckSubmitRawFrame']
        FCB_CheckSubmitRawFrame.restype = ctypes.c_uint32
        FCB_CheckSubmitRawFrame.argtypes = [ctypes.c_uint32, ctypes.c_uint]

        # FCB_SubmitRawFrame(FCBDeviceHandle deviceHandle, const void* frameData, unsigned int frameDataSize, unsigned int* frameId);
        global FCB_SubmitRawFrame
        FCB_SubmitRawFrame = _lib['FCB_SubmitRawFrame']
        FCB_SubmitRawFrame.restype = ctypes.c_uint32
        FCB_SubmitRawFrame.argtypes = [ctypes.c_uint32, ctypes.c_void_p, ctypes.c_uint, ctypes.POINTER(ctypes.c_uint)]

        # FCB_SendRawFrames(FCBDeviceHandle deviceHandle);
        global FCB_SendRawFrames
        FCB_SendRawFrames = _lib['FCB_SendRawFrames']
        FCB_SendRawFrames.restype = ctypes.c_uint32
        FCB_SendRawFrames.argtypes = [ctypes.c_uint32]

        # FCB_WaitForRawTxCompletion(FCBDeviceHandle deviceHandle, unsigned int frameId, TLTimeoutInterval waitTimeout);
        global FCB_WaitForRawTxCompletion
        FCB_WaitForRawTxCompletion = _lib['FCB_WaitForRawTxCompletion']
        FCB_WaitForRawTxCompletion.restype = ctypes.c_uint32
        FCB_WaitForRawTxCompletion.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.c_uint]

        # FCB_GetRawTxPTPTimestamp(FCBDeviceHandle deviceHandle, unsigned int frameId, long long* seconds, unsigned int* nanoseconds);
        global FCB_GetRawTxPTPTimestamp
        FCB_GetRawTxPTPTimestamp = _lib['FCB_GetRawTxPTPTimestamp']
        FCB_GetRawTxPTPTimestamp.restype = ctypes.c_uint32
        FCB_GetRawTxPTPTimestamp.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.POINTER(ctypes.c_uint64), ctypes.POINTER(ctypes.c_uint)]

        # FCB_OpenRawRxDevice(FCBEnumerationHandle enumerationHandle, unsigned int deviceIndex, unsigned int bufferCount, FCBRawHeaderType headerType, FCBDeviceHandle* deviceHandle, unsigned int flags);
        global FCB_OpenRawRxDevice
        FCB_OpenRawRxDevice = _lib['FCB_OpenRawRxDevice']
        FCB_OpenRawRxDevice.restype = ctypes.c_uint32
        FCB_OpenRawRxDevice.argtypes = [ctypes.c_uint32, ctypes.c_uint, ctypes.c_uint, ctypes.c_uint, ctypes.POINTER(ctypes.c_uint), ctypes.c_uint]

        # FCB_EnableRawRx(FCBDeviceHandle deviceHandle);
        global FCB_EnableRawRx
        FCB_EnableRawRx = _lib['FCB_EnableRawRx']
        FCB_EnableRawRx.restype = ctypes.c_uint32
        FCB_EnableRawRx.argtypes = [ctypes.c_uint32]

        # FCB_DisableRawRx(FCBDeviceHandle deviceHandle);
        global FCB_DisableRawRx
        FCB_DisableRawRx = _lib['FCB_DisableRawRx']
        FCB_DisableRawRx.restype = ctypes.c_uint32
        FCB_DisableRawRx.argtypes = [ctypes.c_uint32]
        
        # FCB_WaitForRawFrame(FCBDeviceHandle deviceHandle, TLTimeoutInterval waitTimeout);
        global FCB_WaitForRawFrame
        FCB_WaitForRawFrame = _lib['FCB_WaitForRawFrame']
        FCB_WaitForRawFrame.restype = ctypes.c_uint32
        FCB_WaitForRawFrame.argtypes = [ctypes.c_uint32, ctypes.c_uint]

        # FCB_GetReceivedRawFrame(FCBDeviceHandle deviceHandle, FCBRawFrameType* frameType, void* buffer, unsigned int bufferSize, unsigned int* bytesReturned);
        global FCB_GetReceivedRawFrame
        FCB_GetReceivedRawFrame = _lib['FCB_GetReceivedRawFrame']
        FCB_GetReceivedRawFrame.restype = ctypes.c_uint32
        FCB_GetReceivedRawFrame.argtypes = [ctypes.c_uint32, ctypes.POINTER(ctypes.c_uint), ctypes.c_void_p, ctypes.c_uint, ctypes.POINTER(ctypes.c_uint)]

        # FCB_SpiTransaction(FCBDeviceHandle deviceHandle, unsigned int chipSelectId, const void* writeData, unsigned int bytesToWrite, void* readBuffer, unsigned int bytesToRead, unsigned int flags);
        global FCB_SpiTransaction
        FCB_SpiTransaction = _lib['FCB_SpiTransaction']
        FCB_SpiTransaction.restype = ctypes.c_uint32
        FCB_SpiTransaction.argtypes = [ctypes.c_uint32, ctypes.c_uint32, ctypes.c_void_p, ctypes.c_uint, ctypes.c_void_p, ctypes.c_uint, ctypes.c_uint]


        # --------------- Private Test Functions -----------------------------
        # FCT_I2cTransaction(FCBDeviceHandle deviceHandle,
        #                       unsigned char i2cBus,
        #                       unsigned int i2cSpeed,
        #                       unsigned char i2cSlaveAddress,
        #                       const void* writeData,
        #                       unsigned int bytesToWrite,
        #                       void* readBuffer,
        #                       unsigned int bytesToRead,
        #                       int timeoutInterval);

        global FCT_I2cTransaction
        FCT_I2cTransaction = _lib['FCT_I2cTransaction']
        FCT_I2cTransaction.restype = ctypes.c_uint32
        FCT_I2cTransaction.argtypes = [ctypes.c_uint32, ctypes.c_ubyte, ctypes.c_uint, ctypes.c_ubyte, ctypes.c_void_p, ctypes.c_uint, ctypes.c_void_p, ctypes.c_uint, ctypes.c_uint]
       
        # FCT_ExecuteCommand(FCBDeviceHandle deviceHandle,
        #                       unsigned char messageCode,
        #                       unsigned char subsystem,
        #                       unsigned char instance,
        #                       const void* commandPayload,
        #                       unsigned int commandPayloadLength,
        #                       void* responseBuffer,
        #                       unsigned int responseBufferSize,
        #                       unsigned char* responseCode,
        #                       unsigned int* responseLength,
        #                       unsigned int* responseStatusCode,
        #                       TLTimeoutInterval responseTimeout);
                                       
        global FCT_ExecuteCommand
        FCT_ExecuteCommand = _lib['FCT_ExecuteCommand']
        FCT_ExecuteCommand.restype = ctypes.c_uint32
        FCT_ExecuteCommand.argtypes = [ctypes.c_uint32, ctypes.c_ubyte, ctypes.c_ubyte, ctypes.c_ubyte, ctypes.c_void_p, ctypes.c_uint, ctypes.c_void_p, ctypes.c_uint, ctypes.POINTER(ctypes.c_char), ctypes.POINTER(ctypes.c_uint), ctypes.POINTER(ctypes.c_uint), ctypes.c_int]
       
                
    except AttributeError as err:
        error_msg = err.args[0]
    else:
        error_msg = ''
    finally:
        if error_msg:
            raise LibInitError(error_msg, _lib._name)     
        
        
def is_initialized():
    """Return the initialization status of this module.
    
    Returns:
        `Boolean` -- Module and library are fully initialized.

    @ingroup hid2inic_api

    """
    
    global _lib
    
    if _lib:
        # lib has been loaded
        return True

    return False

def get_unichar_size():
    """get size of unicode/char used on this OS-env"""
    global _lib
    
    if _lib:
        if 'FCB_GetUniCharSize' in globals():
            res = FCB_GetUniCharSize()
            return res
    return -1

def get_api_version():
    """returns Tuple (verMajor, verMinor)"""
    
    global _lib
   
    if _lib:
        if 'FCB_GetApiVersion' in globals():
            res = FCB_GetApiVersion()
            ver_major = (res >> 16) & 0xFFFF
            ver_minor = res & 0xFFFF
            return (ver_major, ver_minor)



def check_api_version(ver_major, ver_minor):
    """verify if correct Dll/so is loaded"""
    
    global _lib
    
    if _lib:
        if 'FCB_CheckApiVersion' in globals():
            res = FCB_CheckApiVersion(ver_major, ver_minor)
            if res == 1:
                return True
            return False


def shutdown():
    """disconnect from USB-HID device"""
    
    global _lib
    
    if _lib:
        if 'FCB_Shutdown' in globals():
            FCB_Shutdown()

def register_device_change(func):
    """register callback for device changes: PNP"""
    
    global _lib
    
    if _lib:
        if 'FCB_RegisterDeviceChangeCallback' in globals():
            handle = ctypes.c_uint32()
            res = FCB_RegisterDeviceChangeCallback(ctypes.c_int(FCBDeviceType.FCBDeviceTypeAny), func, 0, ctypes.byref(handle))
            if res == 0:
                return handle.value
    else:
        raise LibInitError(error_msg, _lib._name)

def unregister_device_change(notifyHandle):
    """unregister callback for device changes: PNP"""
    
    global _lib
    
    if _lib:
        if 'FCB_UnregisterDeviceChangeCallback' in globals():
            handle = ctypes.c_uint32()
            res = FCB_UnregisterDeviceChangeCallback(ctypes.c_int(notifyHandle))
            if res == 0:
                return True
    else:
        raise LibInitError(error_msg, _lib._name)

def get_num_devices():
    """enumerates all connected FCBase devices"""
    
    global _lib
    
    if _lib:
        if 'FCB_EnumerateDevices' in globals():
            count = ctypes.c_uint32()
            enumHandle = ctypes.c_uint()
            res = FCB_EnumerateDevices(ctypes.c_int(FCBDeviceType.FCBDeviceTypeAny), ctypes.byref(enumHandle), ctypes.byref(count))
            FCB_CloseEnumeration(enumHandle)
            if res == 0:
                return count.value
            return -1
    else:
        raise LibInitError(error_msg, _lib._name)
        
def open_control(devType, inst):
    """open a given instance and return the correct device handle
    returns deviceHandle if successfull or None in case of failure
    """
    
    global _lib
    global char_type
    global LastResult
    
    if _lib:
        if 'FCB_OpenControlDevice'in globals():
            count = ctypes.c_uint32()
            enumHandle = ctypes.c_uint()
            res = FCB_EnumerateDevices(ctypes.c_int(devType), ctypes.byref(enumHandle), ctypes.byref(count))
            if res == 0 and count.value > inst:
                devHandle=ctypes.c_uint32()
                res = FCB_OpenControlDevice(enumHandle, ctypes.c_uint32(inst), ctypes.byref(devHandle))
                if res == 0:
                    return devHandle
                elif res == 0xee001001: # unsuppoted FW-Version, too old
                    LastResult = -1
                    return None
    return None

def open_raw_tx(devType, inst):
    """open a given Ethernet Raw-TX instance and return the correct device handle

    """
    
    global _lib
    global char_type
    
    if _lib:
        if 'FCB_OpenRawTxDevice'in globals():
            count = ctypes.c_uint32()
            enumHandle = ctypes.c_uint()
            res = FCB_EnumerateDevices(ctypes.c_int(devType), ctypes.byref(enumHandle), ctypes.byref(count))
            if res == 0 and count.value > inst:
                devHandle=ctypes.c_uint32()
                res = FCB_OpenRawTxDevice(enumHandle, ctypes.c_uint32(inst), 0, 0, ctypes.byref(devHandle), 0)
                if res == 0:
                    return devHandle
    return None

def open_raw_rx(devType, inst, headerType, buffers=0):
    """open a given Ethernet Raw-RX instance and return the correct device handle
        - inst which inst of device to open
        - headerType = RawRxHeaderType
        - buffers = number of buffer-pool, default uses ~80 adjustted inside broadway_api
    """
    
    global _lib
    global char_type
    
    if _lib:
        if 'FCB_OpenRawRxDevice'in globals():
            count = ctypes.c_uint32()
            enumHandle = ctypes.c_uint()
            res = FCB_EnumerateDevices(ctypes.c_int(devType), ctypes.byref(enumHandle), ctypes.byref(count))
            if res == 0 and count.value > inst:
                devHandle=ctypes.c_uint32()
                res = FCB_OpenRawRxDevice(enumHandle, ctypes.c_uint32(inst), buffers, headerType, ctypes.byref(devHandle), 0)
                if res == 0:
                    return devHandle
    return None



def close(devHandle):
    """in opposite to open(), this function closes the device"""
    
    global _lib
    
    if _lib:
        if 'FCB_CloseDevice' in globals():
            try:            
                res = FCB_CloseDevice(devHandle)
                return True
            except:
                return False




def check_device_connection(devHandle):
    """verify, if connection to device is active"""
    global _lib
    
    if _lib:
        if 'FCB_CheckDeviceConnection' in globals():
            try:            
                res = FCB_CheckDeviceConnection(devHandle)
                return True
            except:
                return False

def get_device_property_uint(devHandle, devProperty):
    """retrieve value of given enum property like VendorId, Manufacter, ...
    refer to FCBDeviceProperty for all valid constants
    """
    global _lib
    
    if _lib:
        if 'FCB_GetDevicePropertyUint' in globals():
            try:
                propValue = ctypes.c_uint32()            
                res = FCB_GetDevicePropertyUint(devHandle, devProperty, ctypes.byref(propValue))
                return propValue.value
            except:
                return None

def get_device_property_uint64(devHandle, devProperty):
    """retrieve value of given enum property like VendorId, Manufacter, ...
    refer to FCBDeviceProperty for all valid constants
    """
    global _lib
    
    if _lib:
        if 'FCB_GetDevicePropertyUint64' in globals():
            try:
                propValue = ctypes.c_uint64()            
                res = FCB_GetDevicePropertyUint64(devHandle, devProperty, ctypes.byref(propValue))
                return propValue.value
            except:
                return None

def set_device_property_uint(devHandle, devProperty, newValue, flags):
    """set value of given enum property like VendorId, Manufacter, ...
    refer to FCBDeviceProperty for all valid constants
    """
    global _lib
    
    if _lib:
        if 'FCB_SetDevicePropertyUint' in globals():
            try:
                res = FCB_SetDevicePropertyUint(devHandle, devProperty, newValue, flags)
                if res == 0:
                    return True
            except:
                return False
    return False

def set_device_property_uint64(devHandle, devProperty, newValue, flags):
    """set value of given enum property like VendorId, Manufacter, ...
    refer to FCBDeviceProperty for all valid constants
    """
    global _lib
    
    if _lib:
        if 'FCB_SetDevicePropertyUint64' in globals():
            try:
                res = FCB_SetDevicePropertyUint64(devHandle, devProperty, newValue, flags)
                if res == 0:
                    return True
            except:
                return False
        else:
            return False
    return False

# void* buffer, unsigned int bufferSize, unsigned int* bytesReturned);
def get_device_property_binary(devHandle, devProperty):
    """retrieve value of given enum property like VendorId, Manufacter, ...
    refer to FCBDeviceProperty for all valid constants
    result is given as bytearray
    """
    global _lib
    
    if _lib:
        if 'FCB_GetDevicePropertyBinary' in globals():
            try:
                buffer = (ctypes.c_ubyte*1024)()
                bufferPtr = ctypes.cast(buffer, ctypes.c_void_p)
                bytesReturned = ctypes.c_uint32()
                res = FCB_GetDevicePropertyBinary(devHandle, devProperty, bufferPtr, 512, ctypes.byref(bytesReturned))
                if res == 0:
                    resArray = bytearray(buffer)
                    return resArray[0:bytesReturned.value]
            except:
                return None

def set_device_property_binary(devHandle, devProperty, value, flags):
    """set binary value of given enum property ...
    refer to FCBDeviceProperty for all valid constants
    """
    global _lib
    
    if _lib:
        if 'FCB_SetDevicePropertyBinary' in globals():
            try:
                buffer = (ctypes.c_ubyte*len(value))(*value)
                
                bufferPtr = ctypes.cast(buffer, ctypes.c_void_p)
                res = FCB_SetDevicePropertyBinary(devHandle, devProperty, bufferPtr, len(buffer), flags)
                if res == 0:
                    return True
            except:
                return False
    return False

def get_device_property_string(devHandle, devProperty):
    """retrieve value of given enum property like VendorId, Manufacter, ...
    refer to FCBDeviceProperty for all valid constants
    """
    global _lib
    
    if _lib:
        if 'FCB_GetDevicePropertyString' in globals():
            try:
                buffer = (T_UNICHAR*512)()
                bufferPtr = ctypes.cast(buffer, ctypes.c_void_p)

                res = FCB_GetDevicePropertyString(devHandle, devProperty, bufferPtr, 512)
                if res == 0:
                    # TODO: not optimized how to handle/convert c_char/c_wchar to python string!
                    if T_UNICHAR == ctypes.c_wchar:
                        retStr = ''
                        for j in range(512):
                            if buffer[j] == '\x00':
                                break
                            retStr += buffer[j]
                    else:
                        retStr = ''
                        for j in range(512):
                            if buffer[j] == b'\x00':
                                break
                            retStr += buffer[j].decode()
                return retStr
            except:
                return None

def set_device_property_string(devHandle, devProperty, newValue, flags):
    """retrieve value of given enum property like VendorId, Manufacter, ...
    refer to FCBDeviceProperty for all valid constants
    """
    global _lib
    
    if _lib:
        if 'FCB_SetDevicePropertyString' in globals():
            buffer = (T_UNICHAR*(len(newValue)+1))()
            if T_UNICHAR == ctypes.c_wchar:
                for j in range(len(newValue)):
                    buffer[j] = newValue[j]
            else:
                for j in range(len(newValue)):
                    buffer[j] = newValue[j].encode()
            bufferPtr = ctypes.cast(buffer, ctypes.c_void_p)
            try:
                res = FCB_SetDevicePropertyString(devHandle, devProperty, bufferPtr, flags)
                return res
            except:
                return -1

    return -2

# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------
# --------------------Helper functions to build-in properties ---------------------------
# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------


def set_friendly_name(self, friendlyName, flag):
    """change friendly name used to easy identify multiple instances"""
    broadway_api.set_device_property_string(self.__device, broadway_api.FCBDeviceProperty.FCBDevicePropertyFriendlyName, friendlyName, flag)
    # as stick property has changed, we need to update local fcb properties

def get_device_serial_number(devHandle):
    """receive USB serial number of connected broadway device
    Format is Hex-String with leading zeros: 0000000000000000
    """
    resString = ''
    if _lib:
        if 'FCB_GetDevicePropertyString' in globals():
            try:
                buffer = (T_UNICHAR*512)()
                bufferPtr = ctypes.cast(buffer, ctypes.c_void_p)
                res = FCB_GetDevicePropertyString(devHandle, FCBDeviceProperty.FCBDevicePropertyUsbSerialNumberString, bufferPtr, 512)
                if res == 0:
                    # TODO: not optimized how to handle/convert c_char/c_wchar to python string!
                    if T_UNICHAR == ctypes.c_wchar:
                        for j in range(16):
                            resString += '{0}'.format(buffer[j])
                    else:
                        for j in range(16):
                            resString += buffer[j].decode()
                    return resString 
            except:
                return None
    return resString

def get_timestamp(devHandle):
    """receive 64-bit timestamp of device
    """
    binArray = get_device_property_binary(devHandle, FCBDeviceProperty.FCBDevicePropertyPTPTimestamp)
    if binArray == None:
        return (0 , 0)
    res = (int.from_bytes(binArray[0:8], 'little'), int.from_bytes(binArray[8:12], 'little'))
    return res

def init_timestamp(devHandle, sec, nanosec):
    binArray = sec.to_bytes(8, byteorder='little') + nanosec.to_bytes(4, byteorder='little')
    return set_device_property_binary(devHandle, FCBDeviceProperty.FCBDevicePropertyPTPTimestampInit, binArray, 0)

def update_timestamp(devHandle, sec, nanosec):
    binArray = sec.to_bytes(8, byteorder='little') + nanosec.to_bytes(4, byteorder='little')
    return set_device_property_binary(devHandle, FCBDeviceProperty.FCBDevicePropertyPTPTimestampUpdate, binArray, 0)

def adjust_ptp_clock(devHandle, factor):
    """fix-floating point with 1 digit integer, rest is float
    0x80000000 = 1
    If high-bit is clear, clock moves slower
    If high-bit is set, clock moves faster
    0x90000000 = 1.125
    """
    return set_device_property_uint(devHandle, FCBDeviceProperty.FCBDevicePropertyPTPTimestampClockSpeed, factor, 0)

def read_phy_register(devHandle, registerAddress):
    """direct access to Phy-Chip on Add-On Module"""
    if _lib:
        if 'PHY_ReadRegister' in globals():
            try:
                regValue = ctypes.c_uint32()
                res = PHY_ReadRegister(devHandle, registerAddress, ctypes.byref(regValue))
                if res == 0:
                    return regValue.value
                else:
                    return -1
            except:
                return -2
    return -3

def write_phy_register(devHandle, registerAddress, registerValue):
    """direct access to Phy-Chip on Add-On Module"""
    if _lib:
        if 'PHY_WriteRegister' in globals():
            try:
                res = PHY_WriteRegister(devHandle, registerAddress, registerValue)
                if res == 0:
                    return 0
                else:
                    return -1
            except:
                return -2
    return -3

def read_phy_register_ex(devHandle, phyAddress, registerAddress):
    """direct access to Phy-Chip on Add-On Module"""
    if _lib:
        if 'PHY_ReadRegisterEx' in globals():
            try:
                regValue = ctypes.c_uint32()
                res = PHY_ReadRegisterEx(devHandle, phyAddress, registerAddress, ctypes.byref(regValue))
                if res == 0:
                    return regValue.value
                else:
                    return -1
            except:
                return -2
    return -3

def write_phy_register_ex(devHandle, phyAddress, registerAddress, registerValue):
    """direct access to Phy-Chip on Add-On Module"""
    if _lib:
        if 'PHY_WriteRegisterEx' in globals():
            try:
                res = PHY_WriteRegisterEx(devHandle, phyAddress, registerAddress, registerValue)
                if res == 0:
                    return 0
                else:
                    return -1
            except:
                return -2
    return -3

def read_phy_register45_ex(devHandle, portAddress, deviceAddress, registerAddress):
    """direct read access to Phy-Chip using clause45 SMI"""
    if _lib:
        if 'PHY_ReadRegister45Ex' in globals():
            try:
                regValue = ctypes.c_uint32()
                res = PHY_ReadRegister45Ex(devHandle, portAddress, deviceAddress, registerAddress, ctypes.byref(regValue))
                if res == 0:
                    return regValue.value
                else:
                    return -1
            except:
                return -2
    return -3 

def write_phy_register45_ex(devHandle, portAddress, deviceAddress, registerAddress, registerValue):
    """direct write access to Phy-Chip using clause45 SMI"""
    if _lib:
        if 'PHY_WriteRegister45Ex' in globals():
            try:
                regValue = ctypes.c_uint32()
                res = PHY_WriteRegister45Ex(devHandle, portAddress, deviceAddress, registerAddress, registerValue)
                if res == 0:
                    return 0
                else:
                    return -1
            except:
                return -2
    return -3 

def read_phy_multregisters45_ex(devHandle, portAddress, deviceAddress, registerAddress, numberOfRegisters):
    """direct read-mult access to Phy-Chip using clause45 SMI"""
    if _lib:
        if 'PHY_ReadMultRegisters45Ex' in globals():
            try:
                inBuffer = (ctypes.c_uint*64)()
                inBufferPtr = ctypes.cast(inBuffer, ctypes.c_void_p)
                
                res = PHY_ReadMultRegisters45Ex(devHandle, portAddress, deviceAddress, registerAddress, numberOfRegisters, inBufferPtr)
                if res == 0:
                    resList = []
                    for j in range(numberOfRegisters):
                        resList.append(inBuffer[j])
                    return resList
                else:
                    return -1
            except:
                return -2
    return -3 


def set_pin(devHandle, pin, boolState):
    """Change state of given pin"""
    if _lib:
        if 'FCB_SetPin' in globals():
            try:
                if boolState:
                    pinValue = ctypes.c_uint32(1)
                else:
                    pinValue = ctypes.c_uint32(0)
                res = FCB_SetPin(devHandle, pin, pinValue)
                return True
            except:
                return False

def query_pin(devHandle, pin):
    """Read current state of given pin"""
    if _lib:
        if 'FCB_QueryPin' in globals():
            try:
                pinValue = ctypes.c_uint32()
                res = FCB_QueryPin(devHandle, pin, ctypes.byref(pinValue))
                if res == 0:
                    if pinValue.value > 0:
                        return True
                    else:
                        return False
                else:
                    return None
            except:
                return None

def get_led_mode(devHandle, led):
    """Read current settings for given led"""
    if _lib:
        if 'FCB_GetLedMode' in globals():
            try:
                mode = ctypes.c_uint()
                state = ctypes.c_uint()
                onTime = ctypes.c_uint()
                offTime = ctypes.c_uint()
                res = FCB_GetLedMode(devHandle, led, ctypes.byref(mode), ctypes.byref(state), ctypes.byref(onTime), ctypes.byref(offTime))
                if res == 0:
                    return (led, mode.value, state.value, onTime.value, offTime.value)
            except:
                return ()
    return ()

def set_led_mode(devHandle, led, mode, state, onTime, offTime):
    """Set current settings for given led"""
    if _lib:
        if 'FCB_SetLedMode' in globals():
            try:
                res = FCB_SetLedMode(devHandle, led, mode,  state, onTime, offTime)
                if res == 0:
                    return True
            except:
                return False
    return False

def trigger_wakeup(devHandle):
    """Trigger local wakeup via WAKE_IN_OUT pin"""
    res = set_device_property_uint(devHandle, FCBDeviceProperty.FCBDevicePropertyPhyWake, 1, 0)
    if res == 0:
        return True

    return False

# functions for Ext.Trigger Input
def config_ext_trigger(devHandle, source, mode):
    """config mode for external trigger pin (JTAG-connector UART.Rx or FC401 SPI.MOSI)"""
    binArray = source.to_bytes(1, byteorder='little') + mode.to_bytes(1, byteorder='little')
    return set_device_property_binary(devHandle, FCBDeviceProperty.FCBPropertyExternalTriggerConfig, binArray, 0)

def get_ext_trigger_status(devHandle):
    """get current status of external trigger
    returns tuple for (source, mode)
    source: 1 TRIG_PIN1
    mode: 1 falling edge to GND
    """
    binArray = get_device_property_binary(devHandle, FCBDeviceProperty.FCBPropertyExternalTriggerStatus)
    if binArray == None:
        return (0 , 0)

    numTriggers = int.from_bytes(binArray[0:4], 'little')
    res = []
    for cnt in range(numTriggers):
        res.append((int.from_bytes(binArray[4+cnt*2:4+cnt*2+1], 'little'), int.from_bytes(binArray[4+cnt*2+1:4+cnt*2+2], 'little')))
    
    return res


def get_ext_trigger_out_status(devHandle):
    binArray = get_device_property_binary(devHandle, FCBDeviceProperty.FCBPropertyTriggerOutputStatus)

    if binArray == None:
        return [{'Event': -1, 'Output': -1, 'Mode': -1, 'EtherTypes': [], 'PulseWidth': -1}]
    else:
        if len(binArray) == (8*24+4):    # 8 Outputs 24Bytes each
            numOutputs = int.from_bytes(binArray[0:4], 'little')
            res = []
            for j in range(numOutputs):
                numEtherTypes = int.from_bytes(binArray[4+j*24+3:4+j*24+4], 'little')
                etList = []
                for et in range(numEtherTypes):
                    etList.append(int.from_bytes(binArray[4+j*24+4+et*2:4+j*24+4+et*2+2], 'little'))

                res.append({'Event': int.from_bytes(binArray[4+j*24:4+j*24+1], 'little'),
                   'Output': int.from_bytes(binArray[4+j*24+1:4+j*24+2], 'little'),
                   'Mode': int.from_bytes(binArray[4+j*24+2:4+j*24+3], 'little'),
                   'EtherTypes': etList,
                   'PulseWidth': int.from_bytes(binArray[4+j*24+20:4+j*24+24], 'little'),
                    })

    return res

def config_ext_trigger_out(devHandle, event, output, mode, etherTypeList, pulseWidth):
    numEtherTypes = len(etherTypeList)
    dummy = 0xFFFF
    binArray = event.to_bytes(1, byteorder='little') + output.to_bytes(1, byteorder='little') + mode.to_bytes(1, byteorder='little') + numEtherTypes.to_bytes(1, byteorder='little')

    for j in range(8):
        if (j >= numEtherTypes):
            binArray += dummy.to_bytes(2, byteorder='little')
        else:
            binArray += etherTypeList[j].to_bytes(2, byteorder='little')

    binArray += pulseWidth.to_bytes(4, byteorder='little')

    return set_device_property_binary(devHandle, FCBDeviceProperty.FCBPropertyTriggerOutputConfig, binArray, 0)

def force_trigger_out(devHandle, output):
    res = set_device_property_uint(devHandle, FCBDeviceProperty.FCBPropertyForceTrigger, output, 0)


# functions for 10Base-T1L distance control

def set_long_distance_support(devHandle, enabled):
    """enable/disable long-distance support"""
    if enabled:
        setVal = 1
    else:
        setVal = 0
    res = set_device_property_uint(devHandle, FCBDeviceProperty.FCBPropertySupportLongDistance, setVal, 0)
    if res == 0:
        return True

    return False

def get_long_distance_support(devHandle):
    """return current status of long-distance support"""
    res = get_device_property_uint(devHandle, FCBDeviceProperty.FCBPropertySupportLongDistance)
    
    return res

def enable_long_distance(devHandle, enable, flag):
    """enable/disable long-distance mode
    use flag = 1 to store persistent in Stick EEPROM.
    If store, Stick will startup in stored mode: long or short. Default is LongDistance
    """
    if enable:
        setVal = 1
    else:
        setVal = 0
    res = set_device_property_uint(devHandle, FCBDeviceProperty.FCBPropertyEnableLongDistance, setVal, flag)
    return res

def get_long_distance_mode(devHandle):
    """get current long-distance mode"""
    res = get_device_property_uint(devHandle, FCBDeviceProperty.FCBPropertyEnableLongDistance)
    return res

def set_tx_frame_min_interval(devHandle, minInterval):
    """set min.tx interval in nano-seconds to add pause between sent frames"""
    res = set_device_property_uint64(devHandle, FCBDeviceProperty.FCBPropetyMinTxFrameInterval, minInterval, 0)
    return res

def get_tx_frame_min_interval(devHandle):
    """get current min.tx interval in nano-seconds"""
    res = get_device_property_uint64(devHandle, FCBDeviceProperty.FCBPropetyMinTxFrameInterval)
    return res


def set_smi_passive(devHandle, state):
    """enables to set MDIO pins high-ohmig"""
    res = set_device_property_uint(devHandle, FCBDeviceProperty.FCBPropertySmiInterfacePassive, state, 0)
    return res

def get_smi_passive(devHandle):
    """enables to set MDIO pins high-ohmig"""
    res = get_device_property_uint(devHandle, FCBDeviceProperty.FCBPropertySmiInterfacePassive)
    return res

def set_spi_passive(devHandle, state):
    """enables to set all SPI related pins high-ohmig"""
    res = set_device_property_uint(devHandle, FCBDeviceProperty.FCBPropertySpiInterfacePassive, state, 0)
    return res

def get_spi_passive(devHandle):
    """enables to set MDIO pins high-ohmig"""
    res = get_device_property_uint(devHandle, FCBDeviceProperty.FCBPropertySpiInterfacePassive)
    return res

#
# --------------- Raw Tx functions
#
def wait_for_tx(devHandle, timeout):
    """wait until given level is reached
    
    returns True, if slot get's free in given timeout delay
    return False, if slot get's not freed
    """
    if _lib:
        if 'FCB_WaitForTxResources' in globals():
            res = FCB_WaitForTxResources(devHandle, FCBTxResources.FCBTxResourcesOneHalf, timeout)
            if res == 0:
                return True
            return False

    return None

def check_submit_raw_frame(devHandle, frameDataSize):
    """Checks whether a frame with frameDataSize bytes will fit in the current USB transfer buffer
    
    returns True, if slot with given size is free
    return False, if no buffer available with given size, try later
    """
    if _lib:
        if 'FCB_CheckSubmitRawFrame' in globals():
            res = FCB_CheckSubmitRawFrame(devHandle, frameDataSize)
            if res == 0:
                return True

            return False

    return None 		


def submit_raw_frame(devHandle, data):
    """submits raw frame to buffer and return frameId for processing control"""
    if _lib:
        if 'FCB_SubmitRawFrame' in globals():
            if len(data) > 0:
                bytesToSubmit = ctypes.c_uint(len(data))
                outBufferPtr = ctypes.cast(data, ctypes.POINTER(ctypes.c_void_p))
            else:
                outBufferPtr = ctypes.cast(None, ctypes.c_void_p)
                bytesToSubmit = 0
            frameId = ctypes.c_uint()
                
            res = FCB_SubmitRawFrame(devHandle, outBufferPtr, bytesToSubmit, ctypes.byref(frameId))
            
            if res == 0:
                return frameId.value
    return -1

def send_raw_frames(devHandle):
    """trigger real tranmission of all submitted frames (see above)"""
    if _lib:
        if 'FCB_SendRawFrames' in globals():
            res = FCB_SendRawFrames(devHandle)
            
            if res == 0:
                return True

    return False

def wait_for_raw_tx_completion(devHandle, frameId, timeout):
    """function enables to wait for trigger when frame is really send out"""
    if _lib:
        if 'FCB_WaitForRawTxCompletion' in globals():
            res = FCB_WaitForRawTxCompletion(devHandle, frameId, timeout)
            if res == 0:
                return True

    return False

def get_raw_tx_timestamp(devHandle, frameId):
    """function reads TX-Timestamp when given frameid was sent
    reult is tuple with (sec, nanosec)
    """
    if _lib:
        if 'FCB_GetRawTxPTPTimestamp' in globals():
            sec = ctypes.c_uint64()
            nanosec = ctypes.c_uint()
            res = FCB_GetRawTxPTPTimestamp(devHandle, frameId, ctypes.byref(sec), ctypes.byref(nanosec))
            if res == 0:
                return (sec.value, nanosec.value)

    return (-1, -1)

#
# --------------- Raw Rx functions
#

def enable_raw_rx(devHandle):
    """function enables to receive raw frames"""
    if _lib:
        if 'FCB_EnableRawRx' in globals():
            res = FCB_EnableRawRx(devHandle)
            if res == 0:
                return True

    return False

def disable_raw_rx(devHandle):
    """function enables to receive raw frames"""
    if _lib:
        if 'FCB_DisbleRawRx' in globals():
            res = FCB_DisableRawRx(devHandle)
            if res == 0:
                return True

    return False


def wait_raw_rx_frame(devHandle, timeout):
    """function enables to receive raw frames"""
    if _lib:
        if 'FCB_WaitForRawFrame' in globals():
            res = FCB_WaitForRawFrame(devHandle, timeout)
            if res == 0:
                return True

    return False

def get_raw_rx_frame(devHandle):
    """Gets the data of a received raw frame.
    returns FrameData as byte-buffer to raw frame
    """
    if _lib:
        if 'FCB_GetReceivedRawFrame' in globals():
            inBuffer = (ctypes.c_ubyte*2048)()
            inBufferPtr = ctypes.cast(inBuffer, ctypes.c_void_p)
            frameType = ctypes.c_uint()
            bytesReturned = ctypes.c_uint()

            res = FCB_GetReceivedRawFrame(devHandle, ctypes.byref(frameType), inBufferPtr, 2048, ctypes.byref(bytesReturned))
            if res == 0 and frameType.value == 1:
                if bytesReturned.value > 0:
                    readFrameData = bytes(inBuffer[0:bytesReturned.value])
                else:
                    readFrameData = None
                return readFrameData

    return None

def get_raw_rx_frame_ext(devHandle):
    """Gets the data of a received raw frame.
    return tuple with (frameType, EthRawData)    
    """
    if _lib:
        if 'FCB_GetReceivedRawFrame' in globals():
            inBuffer = (ctypes.c_ubyte*2048)()
            inBufferPtr = ctypes.cast(inBuffer, ctypes.c_void_p)
            frameType = ctypes.c_uint()
            bytesReturned = ctypes.c_uint()

            res = FCB_GetReceivedRawFrame(devHandle, ctypes.byref(frameType), inBufferPtr, 2048, ctypes.byref(bytesReturned))
            if res == 0:
                if bytesReturned.value > 0:
                    readFrameData = bytes(inBuffer[0:bytesReturned.value])
                else:
                    readFrameData = None
                return (frameType.value, readFrameData)

    return None


def enable_idle_frames(devHandle, interval):
    """Set/enable idle frames on raw-rx stream"""
    res = set_device_property_uint(devHandle, FCBDeviceProperty.FCBDevicePropertyFrameInterval, interval, 0)
    if res == 0:
        return True

    return False

def get_frame_stats(devHandle):
    """get frame statistics for ok/Err/CRC/... """

    binArray = get_device_property_binary(devHandle, FCBDeviceProperty.FCBPropertyQueryFrameStatisticsV1)
    
    stat = FrameStatsV1(binArray)
    
    return stat


def get_frame_stats_reset(devHandle):
    """get frame statistics for ok/Err/CRC/... and clear all counters after read"""
    binArray = get_device_property_binary(devHandle, FCBDeviceProperty.FCBPropertyQueryAndClearFrameStatisticsV1)
    stat = FrameStatsV1(binArray)
    
    return stat

# SPI extension used for NXP SJAxxx access

def spi_transaction(devHandle, cs, writeData, bytesToRead, flags):
    """ Generic SPI transfer function """
    if _lib:
        if 'FCB_SpiTransaction' in globals():
            inBuffer = (ctypes.c_ubyte*4096)()
            inBufferPtr = ctypes.cast(inBuffer, ctypes.c_void_p)
            bytesToSubmit = ctypes.c_uint(len(writeData))
            outBufferPtr = ctypes.cast(writeData, ctypes.POINTER(ctypes.c_void_p))
            bytesReturned = ctypes.c_uint()

            res = FCB_SpiTransaction(devHandle, cs, outBufferPtr, bytesToSubmit, inBufferPtr, bytesToRead, flags)
            if res == 0:
                if bytesToRead > 0:
                    readFrameData = bytes(inBuffer[0:bytesToRead])
                else:
                    readFrameData = None
                return readFrameData

    return None

def spi_set_protocol(devHandle, protocolId):
    """configure correct SPI protocol settings
    currently supported is:
    SJA1105: CPHA=1, CPOL=0
    SJA1110: CPHA=0, CPOL=0"""

    if protocolId == FCBSpiProtocol.SJA1105 or protocolId == FCBSpiProtocol.SJA1110:
        res = set_device_property_uint(devHandle, FCBDeviceProperty.FCBPropertySpiProtocol, protocolId, 1)
    else:
        res = -1

    if res == 0:
        return True
    return False

def spi_set_speed(devHandle, speed):
    """ configure correct SPI protocol speed """
    res = set_device_property_uint(devHandle, FCBDeviceProperty.FCBPropertySpiBusSpeed, speed, 1)
    if res == 0:
        return True
    return False
        
#
# -------- test functions
#

def i2c_transaction(devHandle, i2cBus, i2cSpeed, i2cSlaveAddress, writeData, bytesToRead, timeout):
    """ TEST-Only: Generic I2C-Access
    expects writeData as binary string b''
    """
    readData = b''
    if _lib:
        if 'FCT_I2cTransaction' in globals():
            try:
                if bytesToRead > 0:
                    inBuffer = (ctypes.c_ubyte*bytesToRead)()
                    inBufferPtr = ctypes.cast(inBuffer, ctypes.c_void_p)
                else:
                    inBufferPtr = ctypes.cast(None, ctypes.c_void_p)
                if len(writeData) > 0:
                    bytesToWrite = ctypes.c_uint(len(writeData))
                    outBufferPtr = ctypes.cast(writeData, ctypes.POINTER(ctypes.c_void_p))
                else:
                    outBufferPtr = ctypes.cast(None, ctypes.c_void_p)
                    bytesToWrite = 0
                res = FCT_I2cTransaction(devHandle, i2cBus, i2cSpeed, i2cSlaveAddress, outBufferPtr, bytesToWrite, inBufferPtr, bytesToRead, timeout)
                if res == 0 and bytesToRead > 0:
                    readData = bytes(inBuffer)
            except:
                pass
    return readData

def execute_command(devHandle, cmd, subsystem, payloadData, timeout):
    """TEST-Only: Toggle Pins"""
    if _lib:
        if 'FCT_ExecuteCommand' in globals():
            bytesToReadMax = 512
            respCode = ctypes.c_char()
            respLength = ctypes.c_uint()
            respStatus = ctypes.c_uint()
            inBuffer = (ctypes.c_ubyte*bytesToReadMax)()
            inBufferPtr = ctypes.cast(inBuffer, ctypes.c_void_p)

            bytesToWrite = ctypes.c_uint(len(payloadData))
            outBufferPtr = ctypes.cast(payloadData, ctypes.POINTER(ctypes.c_void_p))
                
            res = FCT_ExecuteCommand(devHandle, cmd, subsystem, 0, outBufferPtr, bytesToWrite, inBufferPtr, bytesToReadMax, ctypes.byref(respCode), ctypes.byref(respLength), ctypes.byref(respStatus), timeout)
            if res == 0:
                if respLength.value > 0:
                    readData = bytes(inBuffer)
                else:
                    readData = None
            
    return readData

#
# wrapper helper functions
#
    
def parse_result_code(result_code):
    """Parse the result code of a library function and return a string
    describing the error.
        
    Args:
        result_code: `Integer` -- The result code provided by a library function.

    Returns:
        `String` -- Description of the error to the given @a result_code.

    @ingroup hid2inic_api

    """
    if result_code == NO_ERROR:
        return 'No error occurred!'
    elif result_code == NOT_INITIALIZED_ERROR:
        return 'Library not initialized!'
    elif result_code == DEVICE_ERROR:
        return 'Could not get device!'
    elif result_code == INVALID_ARGUMENT_ERROR:
        return 'Invalid argument / Out of range!'
    elif result_code == POINTER_ERROR:
        return 'C pointer error!'
    elif result_code == OUT_OF_MEMORY_ERROR:
        return 'Out of memory!'
    elif result_code == NOT_IMPLEMENTED:
        return 'Function not implemented yet!'
    else:
        return 'Unknown / Unexpected Error!'
    
class Error(Exception):
    """Baseclass for errors in this module.

    @ingroup hid2inic_api
    
    """
    
    pass
    
    
class LibLoadError(Error):
    """Exception raised if library could not be found.

    Attributes:
        path: `String` -- Path of the library
        lib: `String` -- Filename of the library

    @ingroup hid2inic_api
        
    """
    
    path = ""
    lib = ""

    def __init__(self, lib_path):
        """Initialize the instanze.
        
        Args:
            lib_path: `String` -- Full path (including filename) of the library.
        
        """
        split_path = os.path.split(lib_path)
        self.path = split_path[0]
        self.lib = split_path[1]
    
    def __str__(self):
        """Return the string representative of this exception.
        
        Format:
            \"Could not find [@ref lib] in [@ref path]\"
            
        """
        if not self.path:
            return str('Could not find \'' + self.lib + '\'')
        else:
            return str('Could not find \'' + self.lib + '\' in \'' + self.path + '\'')
            

class LibInitError(Error):
    """Exception raised if an element could not be found in library.

    Attributes:
        element: `String` -- Element that failed to initialize / Error message
        lib: `String` -- Full path (including filename) of the library

    @ingroup hid2inic_api
    
    """
    
    element = ""
    lib = ""

    def __init__(self, element, lib):
        """Initialize the instanze.
        
        Args:
            element: `String` -- Element that failed to initialize / Error message.
            lib: `String` -- Full path (including filename) of the library.
        
        """
        self.element = element
        self.lib = lib
    
    def __str__(self):
        """Return the string representative of this exception.
        
        Format:
            \"Symbol [@ref element] not found in [@ref lib]\"
            
        """
        return str(self.element + ' in \'' + self.lib + '\'')
       
        
if __name__ == '__main__':
    # print documentation and exit
    __load_library(lib_path='')
    exit(__doc__)
    

# end of python wrapper
        
        
        
        
        
        
        
        
        
        
        
        
