#!python3
# multi_raw_receiver.py
#
# This file is part of python wrapper for the BROADWAY2 package.
#
# MIT License

# Copyright (c) 2013-2018 FibreCode GmbH <info@fibrecode.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import threading, queue

from broadway2 import broadway_api

class MultiRawReceiverThread(threading.Thread):
    def __init__(self, q, e):
        """initialize receiver instance for multiple FC611/FC612"""
        threading.Thread.__init__(self) 
        self.devRx = []
        self.timeStampOffset = {}
        self.lutInstName = {}    # lut to get name for given instance
        self.listInstance = []
        self.outQ = q
        self.stopEvent = e

    def add(self, adapter, devRx):
        """append Raw-Rx device to receiver"""
        self.devRx.append(devRx)
        self.listInstance.append(adapter.Inst)
        self.lutInstName[adapter.Inst] = adapter.FriendlyName

    def remove(self, dev):
        """remove devRx from receiver list"""
        try:
            idx = self.devRx.index(dev)
            self.devRx.remove(dev)
            self.lutInstName.pop(self.lutInstName[dev])
            self.listInstance.remove(self.listInstance[idx])
        except:
            pass

    def enumerate(self):
        for d in self.devRx:
            yield d

    @property
    def Count(self):
        """returns number of current devs added"""
        return len(self.devRx)

    def rx_enable(self):
        """start all receiving devices"""
        cnt = len(self.devRx)
        for d in self.devRx:
            res = broadway_api.enable_raw_rx(d)
            if res == True:
                cnt -= 1
        return cnt == 0

    def rx_disable(self):
        """stop all receiving devices"""
        cnt = len(self.devRx)
        for d in self.devRx:
            res = broadway_api.disable_raw_rx(d)
            if res:
                cnt -= 1
        return cnt == 0

    def reset_time(self):
        """reset all rx-devices to zero timestamp"""
        for d in self.devRx:
            broadway_api.init_timestamp(d, 0, 0)

    def run(self):
        tm = 0    
            
        idx = 0
        idxLast = len(self.devRx)-1
        while True:
            if self.stopEvent.wait(0):
                break

            res = broadway_api.wait_raw_rx_frame(self.devRx[idx], 20)
            
            if res:
                res = broadway_api.get_raw_rx_frame(self.devRx[idx])
                if res != None:
                    self.outQ.put((self.lutInstName[self.listInstance[idx]], res))
            idx += 1
            if idx > idxLast:
                idx = 0

