# spi_access.py
#
# This file is part of python wrapper for the BROADWAY2 package.
#
# MIT License

# Copyright (c) 2013-2019 FibreCode GmbH <info@fibrecode.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import struct

from broadway2 import broadway_api, spi_access


class SJAXAccess:
    """implement 32-bit SPI protocol for NXP Switch SJA1105x and SJA1110x
    access to SPI is supported on FC401 only, as 4 SPI lanes are only available on USB Tool-Adapter
    """
    def __init__(self, spiAccess, cs):
        """Default constructor requires connected and configured SPIAccess() instance
        spiAccess has fixed protocol-setup and pre-configured speed
        """
        self.__spiAccess = spiAccess
        self.__cs = cs
        self.__dev = spiAccess.get_device() # current USB-control device to access adapter

    def spi_read_write(self, isWrite, regCount, regAddr, regValues):
        """read/write transaction for SJAXXX 32-bit values"""

        if regCount > 64:
            return None

        if self.__dev == None:
            return None

        if isWrite:
            hdr = 1 << 31
        else:
            hdr = 0

        hdr |= (regCount << 25) & 0x7E000000
        hdr |= (regAddr << 4) & 0x01FFFFF0
            
        txDataBin = hdr.to_bytes(4, byteorder='big')

        if isWrite:
            for j in range(0, regCount):
                txDataBin += regValues[j].to_bytes(4, byteorder='big')
                #txDataBin += struct.pack('>L', regValues[j])
            readCnt = 0
        else:
            readCnt = regCount*4
                
        
        resValue = broadway_api.spi_transaction(self.__dev, self.__cs, txDataBin, readCnt, 0)

        if isWrite:
            return 0
        else:
            if regCount == 1:
                return int.from_bytes(resValue, byteorder='big')
            else:
                res = []
                for c in range(0,regCount):
                    val = int.from_bytes(resValue[c*4:c*4+4], byteorder='big')
                    res.append(val)
                return res


    def read_reg(self, regAddr):
        """single 32-bit read access"""
        
        hdr = (1 << 25) & 0x7E000000
        hdr |= (regAddr << 4) & 0x01FFFFF0

        res = broadway_api.spi_transaction(self.__dev, self.__cs, hdr.to_bytes(4, byteorder='big'), 4, 0)
        
        return int.from_bytes(res, byteorder='big')

    def read_regs(self, regAddr, regCount):
        """multiple 32-bit read access using selected cs"""
        res = self.spi_read_write(False, regCount, regAddr, [])

        return res


    def write_reg(self, regAddr, regValue):
        """single 32-bit write register access"""
        
        hdr = 1 << 31
        hdr |= (1 << 25) & 0x7E000000
        hdr |= (regAddr << 4) & 0x01FFFFF0

        txDataBin = hdr.to_bytes(4, byteorder='big')
        txDataBin += regValue.to_bytes(4, byteorder='big')
        
        res = broadway_api.spi_transaction(self.__dev, self.__cs, txDataBin, 8, 0)
        
        return res

    def write_regs(self, regAddr, regValues):
        """write a number of 32-bit values from regValues (list of numbers)"""
        
        res = self.spi_read_write(True, len(regValues), regAddr, regValues)
        
        return res

